import json
import uuid
from typing import Any

import pytest

from eventstoredb import Client
from eventstoredb.events import JsonEvent, JsonRecordedEvent, ReadEvent, RecordedEvent
from eventstoredb.exceptions import StreamNotFoundError
from eventstoredb.options import ReadDirection, ReadStreamOptions, StreamPosition

from .utils import json_test_events


def recorded_event_type(x: Any) -> str:
    __tracebackhide__ = True
    if not isinstance(x, ReadEvent):
        pytest.fail(f"not a ReadEvent: {x}")
    if not isinstance(x.event, RecordedEvent):
        pytest.fail(f"does not contain RecordedEvent: {x}")
    else:
        return x.event.type


@pytest.mark.xfail
def test_recorded_event_type_wrong_type() -> None:
    assert recorded_event_type(None) == "Foo"


@pytest.mark.xfail
def test_recorded_event_type_no_event() -> None:
    assert recorded_event_type(ReadEvent()) == "Foo"


async def test_read_stream_json(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test", data=json.dumps({"some": "data"}).encode())],
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
    )
    events = [e async for e in it]

    assert len(events) == 1
    assert isinstance(events[0], ReadEvent)
    event = events[0].event
    assert event is not None
    assert isinstance(event, RecordedEvent)
    assert event.stream_name == stream_name
    assert isinstance(event.id, uuid.UUID)
    assert event.revision == 0
    assert isinstance(event, JsonRecordedEvent)
    assert event.data is not None
    assert event.data == json.dumps({"some": "data"}).encode()
    assert event.metadata is None


@pytest.mark.skip(reason="test not implemented")
async def test_read_stream_binary() -> None:
    pass


async def test_read_stream_from_start(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(from_revision=StreamPosition.START),
    )
    events = [e async for e in it]

    assert len(events) == 4
    assert recorded_event_type(events[0]) == "Test1"
    assert recorded_event_type(events[1]) == "Test2"
    assert recorded_event_type(events[2]) == "Test3"
    assert recorded_event_type(events[3]) == "Test4"


async def test_read_stream_from_revision(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(from_revision=2),
    )
    events = [e async for e in it]

    assert len(events) == 2
    assert recorded_event_type(events[0]) == "Test3"
    assert recorded_event_type(events[1]) == "Test4"


async def test_read_stream_backwards_from_end(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(
            from_revision=StreamPosition.END,
            direction=ReadDirection.BACKWARDS,
        ),
    )
    events = [e async for e in it]

    assert len(events) == 4
    assert recorded_event_type(events[0]) == "Test4"
    assert recorded_event_type(events[1]) == "Test3"
    assert recorded_event_type(events[2]) == "Test2"
    assert recorded_event_type(events[3]) == "Test1"


async def test_read_stream_backwards_from_revision(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(
            from_revision=1,
            direction=ReadDirection.BACKWARDS,
        ),
    )
    events = [e async for e in it]

    assert len(events) == 2
    assert recorded_event_type(events[0]) == "Test2"
    assert recorded_event_type(events[1]) == "Test1"


async def test_read_stream_max_count(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(max_count=2),
    )
    events = [e async for e in it]

    assert len(events) == 2
    assert recorded_event_type(events[0]) == "Test1"
    assert recorded_event_type(events[1]) == "Test2"


async def test_read_stream_max_count_from_revision(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.read_stream(
        stream_name=stream_name,
        options=ReadStreamOptions(from_revision=1, max_count=2),
    )
    events = [e async for e in it]

    assert len(events) == 2
    assert recorded_event_type(events[0]) == "Test2"
    assert recorded_event_type(events[1]) == "Test3"


# NOTE this test currently creates a "task was destroyed"-error
# see: https://github.com/pytest-dev/pytest-asyncio/issues/235
async def test_read_stream_stream_not_exist(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    it = eventstoredb_client.read_stream(stream_name=stream_name)

    with pytest.raises(StreamNotFoundError) as execinfo:
        _ = [e async for e in it]

    assert execinfo.value.stream_name == stream_name


@pytest.mark.skip(reason="test not implemented")
async def test_read_stream_stream_deleted() -> None:
    pass



from __future__ import annotations

from typing import Type
from uuid import UUID

import betterproto

from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.types import (
    NackAction,
    PersistentSubscriptionConfirmation,
    SubscribeToPersistentSubscriptionToStreamOptions,
)
from eventstoredb.events import (
    BinaryRecordedEvent,
    ContentType,
    JsonRecordedEvent,
    PersistentSubscriptionEvent,
)
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier, Uuid
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    ReadReq,
    ReadReqAck,
    ReadReqNack,
    ReadReqNackAction,
    ReadReqOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespReadEvent,
    ReadRespReadEventRecordedEvent,
    ReadRespSubscriptionConfirmation,
)
from eventstoredb.types import AllPosition


def get_original_event_id(event: PersistentSubscriptionEvent) -> UUID | None:
    if event.link:
        return event.link.id
    elif event.event:
        return event.event.id
    else:
        return None


def create_read_request(
    stream_name: str,
    group_name: str,
    options: SubscribeToPersistentSubscriptionToStreamOptions,
) -> ReadReq:
    request_options = ReadReqOptions()
    request_options.stream_identifier = StreamIdentifier(stream_name.encode())
    request_options.group_name = group_name
    request_options.buffer_size = options.buffer_size
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    return ReadReq(options=request_options)


def create_ack_request(
    events: PersistentSubscriptionEvent | list[PersistentSubscriptionEvent],
) -> ReadReq:
    if not isinstance(events, list):
        events = [events]

    ack = ReadReqAck()
    for event in events:
        event_id = get_original_event_id(event)
        if event_id:
            ack.ids.append(Uuid(string=str(event_id)))
    return ReadReq(ack=ack)


def create_nack_request(
    action: NackAction,
    reason: str,
    events: PersistentSubscriptionEvent | list[PersistentSubscriptionEvent],
) -> ReadReq:
    if not isinstance(events, list):
        events = [events]

    nack = ReadReqNack()
    nack.reason = reason

    if action == NackAction.PARK:
        nack.action = ReadReqNackAction.Park
    elif action == NackAction.RETRY:
        nack.action = ReadReqNackAction.Retry
    elif action == NackAction.SKIP:
        nack.action = ReadReqNackAction.Skip
    elif action == NackAction.STOP:
        nack.action = ReadReqNackAction.Stop
    elif action == NackAction.UNKNOWN:
        nack.action = ReadReqNackAction.Unknown

    for event in events:
        event_id = get_original_event_id(event)
        if event_id:
            nack.ids.append(Uuid(string=str(event_id)))
    return ReadReq(nack=nack)


def convert_read_response(
    message: ReadResp,
) -> PersistentSubscriptionConfirmation | PersistentSubscriptionEvent:
    content_type, _ = betterproto.which_one_of(message, "content")
    if content_type == "event":
        return convert_read_response_event(message.event)
    elif content_type == "subscription_confirmation":
        return convert_read_response_confirmation(message.subscription_confirmation)
    else:
        # TODO raise better exception
        raise Exception("shouldnt be here")


def convert_read_response_event(
    message: ReadRespReadEvent,
) -> PersistentSubscriptionEvent:
    event = PersistentSubscriptionEvent()
    if message.event:
        event.event = convert_read_response_recorded_event(message.event)
    if message.link:
        event.link = convert_read_response_recorded_event(message.link)
    if message.commit_position:
        event.commit_position = message.commit_position
    if message.retry_count:
        event.retry_count = message.retry_count
    return event


def convert_read_response_recorded_event(
    message: ReadRespReadEventRecordedEvent,
) -> JsonRecordedEvent | BinaryRecordedEvent:
    stream_name = message.stream_identifier.stream_name.decode()
    id = UUID(message.id.string)
    content_type = ContentType(message.metadata["content-type"])
    position = AllPosition(
        commit_position=message.commit_position,
        prepare_position=message.prepare_position,
    )
    event_class: Type[JsonRecordedEvent] | Type[BinaryRecordedEvent]
    if content_type == ContentType.JSON:
        event_class = JsonRecordedEvent
    else:
        event_class = BinaryRecordedEvent
    return event_class(
        stream_name=stream_name,
        id=id,
        revision=message.stream_revision,
        type=message.metadata["type"],
        content_type=content_type,
        created=int(message.metadata["created"]),
        position=position,
        data=message.data if message.data else None,
        metadata=message.custom_metadata if message.custom_metadata else None,
    )


def convert_read_response_confirmation(
    message: ReadRespSubscriptionConfirmation,
) -> PersistentSubscriptionConfirmation:
    return PersistentSubscriptionConfirmation(id=message.subscription_id)



# flake8: noqa
from eventstoredb.client.append_to_stream.exceptions import (
    AppendToStreamError,
    RevisionMismatchError,
    StreamAlreadyExistsError,
)
from eventstoredb.client.create_persistent_subscription_to_stream.exceptions import (
    PersistentSubscriptionAlreadyExistsError,
    PersistentSubscriptionDroppedError,
    PersistentSubscriptionError,
    PersistentSubscriptionMaxSubscribersReachedError,
    PersistentSubscriptionNotFoundError,
)
from eventstoredb.client.exceptions import (
    ClientException,
    ConnectionStringError,
    ConnectionStringMalformedError,
    ConnectionStringMissingHostError,
    StreamNotFoundError,
)



from __future__ import annotations

import betterproto

from eventstoredb.client.append_to_stream.exceptions import (
    RevisionMismatchError,
    StreamAlreadyExistsError,
)
from eventstoredb.client.append_to_stream.types import (
    AppendExpectedRevision,
    AppendResult,
    AppendToStreamOptions,
)
from eventstoredb.client.exceptions import StreamNotFoundError
from eventstoredb.events import EventData
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier, Uuid
from eventstoredb.generated.event_store.client.streams import (
    AppendReq,
    AppendReqOptions,
    AppendReqProposedMessage,
    AppendResp,
    AppendRespSuccess,
    AppendRespWrongExpectedVersion,
)
from eventstoredb.types import AllPosition, StreamRevision


def create_append_header(stream_name: str, options: AppendToStreamOptions) -> AppendReq:
    request_options = AppendReqOptions()
    request_options.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.expected_revision, StreamRevision):
        request_options.revision = options.expected_revision
    elif options.expected_revision == AppendExpectedRevision.STREAM_EXISTS:
        request_options.stream_exists = Empty()
    elif options.expected_revision == AppendExpectedRevision.NO_STREAM:
        request_options.no_stream = Empty()
    else:
        request_options.any = Empty()

    return AppendReq(options=request_options)


def create_append_request(event_data: EventData) -> AppendReq:
    message = AppendReqProposedMessage()
    message.id = Uuid(string=str(event_data.id))
    message.metadata["type"] = event_data.type
    message.metadata["content-type"] = event_data.content_type
    if event_data.data:
        message.data = event_data.data
    if event_data.metadata:
        message.custom_metadata = event_data.metadata
    return AppendReq(proposed_message=message)


def convert_append_response(stream_name: str, message: AppendResp) -> AppendResult:
    result_type, _ = betterproto.which_one_of(message, "result")
    if result_type == "wrong_expected_version":
        raise convert_wrong_expected_version(
            stream_name=stream_name,
            message=message.wrong_expected_version,
        )
    elif result_type == "success":
        return convert_append_response_success(message=message.success)
    else:
        # TODO raise a more specific exception
        raise Exception("I shouldnt be here")


def convert_append_response_success(message: AppendRespSuccess) -> AppendResult:
    position_type, _ = betterproto.which_one_of(message, "position_option")
    if position_type == "position":
        position = AllPosition(
            commit_position=message.position.commit_position,
            prepare_position=message.position.prepare_position,
        )
    else:
        position = None
    return AppendResult(
        success=True,
        next_expected_revision=message.current_revision,
        position=position,
    )


def convert_wrong_expected_version(
    stream_name: str,
    message: AppendRespWrongExpectedVersion,
) -> StreamNotFoundError | StreamAlreadyExistsError | RevisionMismatchError:
    expected_type, _ = betterproto.which_one_of(message, "expected_revision_option")
    current_type, _ = betterproto.which_one_of(message, "current_revision_option")

    if expected_type == "expected_no_stream":
        return StreamAlreadyExistsError(stream_name=stream_name)
    elif expected_type == "expected_stream_exists":
        return StreamNotFoundError(stream_name=stream_name)
    elif expected_type == "expected_revision" and current_type == "current_no_stream":
        return StreamNotFoundError(stream_name=stream_name)
    elif expected_type == "expected_any" and current_type == "current_no_stream":
        return StreamNotFoundError(stream_name=stream_name)
    elif expected_type == "expected_revision" and current_type == "current_revision":
        return RevisionMismatchError(
            stream_name=stream_name,
            expected_revision=message.expected_revision,
            current_revision=message.current_revision,
        )
    elif expected_type == "expected_any" and current_type == "current_revision":
        return RevisionMismatchError(
            stream_name=stream_name,
            expected_revision=AppendExpectedRevision.ANY,
            current_revision=message.current_revision,
        )
    # FIXME maybe we should raise something like "UnexpectedRuntimeError" here?
    return Exception("SHOULDNT BE HERE")  # type: ignore



import asyncio

from eventstoredb import Client
from eventstoredb.events import JsonEvent
from eventstoredb.options import StreamPosition, SubscribeToStreamOptions

from .utils import Consumer, json_test_events


async def test_subscribe_to_stream_json(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    it = eventstoredb_client.subscribe_to_stream(stream_name=stream_name)

    consumer = Consumer(it)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await consumer.stop(1)

    assert len(consumer.events) == 1


async def test_subscribe_to_stream_from_start(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=StreamPosition.START),
    )

    consumer = Consumer(it)
    await consumer.start()
    await asyncio.sleep(1)

    assert len(consumer.events) == 4

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await consumer.stop(1)

    assert len(consumer.events) == 5


async def test_subscribe_to_stream_from_end(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=StreamPosition.END),
    )

    consumer = Consumer(it)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await consumer.stop(1)

    assert len(consumer.events) == 1


async def test_subscribe_to_stream_from_revision(
    eventstoredb_client: Client,
    stream_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(4),
    )

    it = eventstoredb_client.subscribe_to_stream(
        stream_name=stream_name,
        options=SubscribeToStreamOptions(from_revision=2),
    )

    consumer = Consumer(it)
    await consumer.start()
    await asyncio.sleep(1)

    assert len(consumer.events) == 1

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=JsonEvent(type="Test"),
    )

    await consumer.stop(1)

    assert len(consumer.events) == 2



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from operator import itemgetter
from typing import Any, Callable, Coroutine
from urllib.parse import quote

import requests  # type: ignore

from eventstoredb.events import JsonEvent, ReadEvent
from eventstoredb.subscriptions import PersistentSubscription, Subscription


def json_test_events(amount: int) -> list[JsonEvent]:
    events = []
    for i in range(amount):
        events.append(JsonEvent(type=f"Test{i+1}"))
    return events


@dataclass
class EventstoreHTTP:
    host: str
    port: int

    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}"

    def read_stream(self, stream_name: str, event_id: int | None = None) -> Any:
        url = f"{self.url}/streams/{quote(stream_name)}"
        if event_id is not None:
            url = f"{url}/{event_id}"
        response = requests.get(
            url=url,
            headers={"Accept": "application/vnd.eventstore.atom+json"},
            params={"embed": "TryHarder"},
        )
        data = response.json()
        if event_id is not None:
            return data
        else:
            events = data.get("entries")
            return sorted(events, key=itemgetter("positionEventNumber"))

    def get_persistent_subscriptions(self, stream_name: str | None = None) -> Any:
        url = f"{self.url}/subscriptions"
        if stream_name:
            url = f"{url}/{stream_name}"
        response = requests.get(url=url)
        return response.json()

    def get_persistent_subscription_details(
        self, stream_name: str, group_name: str
    ) -> Any | None:
        url = f"{self.url}/subscriptions/{stream_name}/{group_name}/info"
        response = requests.get(url=url)
        if response.status_code == 200:
            return response.json()
        return None


class Consumer:
    def __init__(
        self,
        it: Subscription | PersistentSubscription,
        on_event: Callable[[Any], Coroutine[Any, Any, None]] | None = None,
    ) -> None:
        self.events: list[ReadEvent] = []
        self._it = it
        self._on_event = on_event
        self._task: asyncio.Task[Any] | None = None
        self._consume_started = asyncio.Event()

    async def start(self) -> None:
        self._task = asyncio.create_task(self._consume())
        await self._consume_started.wait()

    async def stop(self, delay: int | float | None = None) -> None:
        if delay:
            await asyncio.sleep(delay)

        if self._task:
            self._task.cancel()
            try:
                await self._task
            except asyncio.CancelledError:
                pass

    async def run_for(self, duration: int | float) -> None:
        if not self._task:
            await self.start()
        await asyncio.sleep(duration)
        await self.stop()

    async def _consume(self) -> None:
        self._consume_started.set()
        async for event in self._it:
            if self._on_event:
                await self._on_event(event)
            self.events.append(event)



import pytest

from eventstoredb import Client
from eventstoredb.exceptions import PersistentSubscriptionNotFoundError
from eventstoredb.options import (
    CreatePersistentSubscriptionToAllOptions,
    StreamPosition,
    UpdatePersistentSubscriptionToAllOptions,
)

from .utils import EventstoreHTTP


async def test_update_persistent_subscription_to_all(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.START
        ),
    )
    details_before = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details_before is not None
    assert details_before["config"]["startPosition"] == "C:0/P:0"

    await eventstoredb_client.update_persistent_subscription_to_all(
        group_name=group_name,
        options=UpdatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.END
        ),
    )

    details_after = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details_after is not None
    assert details_after["config"]["startPosition"] == "C:-1/P:-1"


async def test_update_persistent_subscription_to_all_raises_if_not_found(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    with pytest.raises(PersistentSubscriptionNotFoundError):
        await eventstoredb_client.update_persistent_subscription_to_all(
            group_name=group_name,
        )



from __future__ import annotations

from dataclasses import dataclass

from eventstoredb.types import ReadDirection, StreamPosition, StreamRevision


@dataclass
class ReadStreamOptions:
    from_revision: StreamRevision | StreamPosition = StreamPosition.START
    max_count: int = 2**64 - 1  # max-uint64
    direction: ReadDirection = ReadDirection.FORWARDS
    resolve_links: bool = False



from dataclasses import dataclass

from eventstoredb.client.create_persistent_subscription_to_all.types import (
    CreatePersistentSubscriptionToAllOptions,
)


@dataclass
class UpdatePersistentSubscriptionToAllOptions(
    CreatePersistentSubscriptionToAllOptions
):
    pass



from __future__ import annotations

from eventstoredb.client.create_persistent_subscription_to_all.grpc import (
    create_persistent_subscription_request_settings,
)
from eventstoredb.client.update_persistent_subscription_to_all.types import (
    UpdatePersistentSubscriptionToAllOptions,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    UpdateReq,
    UpdateReqAllOptions,
    UpdateReqConsumerStrategy,
    UpdateReqOptions,
    UpdateReqPosition,
    UpdateReqSettings,
)
from eventstoredb.types import AllPosition, StreamPosition


def create_update_persistent_subscription_to_all_request(
    group_name: str,
    options: UpdatePersistentSubscriptionToAllOptions,
) -> UpdateReq:
    request_options = UpdateReqOptions()
    request_options.all = UpdateReqAllOptions()
    request_options.group_name = group_name

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = UpdateReqPosition()
        request_options.all.position.commit_position = (
            options.from_position.commit_position
        )
        request_options.all.position.prepare_position = (
            options.from_position.prepare_position
        )
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    request_options.settings = create_persistent_subscription_request_settings(
        settings=options.settings,
        settings_class=UpdateReqSettings,
        consumer_strategy_class=UpdateReqConsumerStrategy,
    )

    return UpdateReq(options=request_options)



from __future__ import annotations


class ConnectionStringError(ValueError):
    def __init__(self, connection_string: str, message: str) -> None:
        self.connection_string = connection_string
        super().__init__(message)


class ConnectionStringMalformedError(ConnectionStringError):
    def __init__(self, connection_string: str) -> None:
        super().__init__(
            connection_string=connection_string,
            message=f"ConnectionString '{connection_string}' is malformed",
        )


class ConnectionStringMissingHostError(ConnectionStringError):
    def __init__(self, connection_string: str) -> None:
        super().__init__(
            connection_string=connection_string,
            message=f"ConnectionString '{connection_string}' is missing host",
        )


class ClientException(Exception):
    ...


class StreamNotFoundError(ClientException):
    def __init__(self, stream_name: str) -> None:
        self.stream_name = stream_name
        super().__init__(f"Stream '{stream_name}' not found")



from __future__ import annotations

from dataclasses import dataclass

from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.types import (
    SubscribeToPersistentSubscriptionToStreamOptions,
)


@dataclass
class SubscribeToPersistentSubscriptionToAllOptions(
    SubscribeToPersistentSubscriptionToStreamOptions
):
    ...



from __future__ import annotations

from dataclasses import dataclass

from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.types import AllPosition, StreamPosition


@dataclass
class SubscribeToAllOptions:
    from_position: AllPosition | StreamPosition = StreamPosition.START
    resolve_links: bool = False
    filter: ExcludeSystemEventsFilter | EventTypeFilter | StreamNameFilter | None = None
    max_search_window: int | None = None
    checkpoint_interval: int = 1


@dataclass
class Checkpoint(AllPosition):
    ...



from __future__ import annotations

from grpclib.exceptions import GRPCError

from eventstoredb.client.create_persistent_subscription_to_stream.grpc import (
    convert_grpc_error_to_exception,
)
from eventstoredb.client.delete_persistent_subscription_to_all.grpc import (
    create_delete_persistent_subscription_to_all_request,
)
from eventstoredb.client.delete_persistent_subscription_to_all.types import (
    DeletePersistentSubscriptionToAllOptions,
)
from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    PersistentSubscriptionsStub,
)


class DeletePersistentSubscriptionToAllMixin(ClientProtocol):
    async def delete_persistent_subscription_to_all(
        self,
        group_name: str,
        options: DeletePersistentSubscriptionToAllOptions | None = None,
    ) -> None:
        if options is None:
            options = DeletePersistentSubscriptionToAllOptions()

        client = PersistentSubscriptionsStub(channel=self.channel)
        request = create_delete_persistent_subscription_to_all_request(
            group_name=group_name,
            options=options,
        )

        try:
            await client.delete(delete_req=request)
        except GRPCError as e:
            raise convert_grpc_error_to_exception(e)



from __future__ import annotations

from eventstoredb.client.read_all.types import ReadAllOptions
from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.streams import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsAllOptions,
    ReadReqOptionsFilterOptions,
    ReadReqOptionsFilterOptionsExpression,
    ReadReqOptionsPosition,
    ReadReqOptionsReadDirection,
    ReadReqOptionsUuidOption,
)
from eventstoredb.types import AllPosition, ReadDirection, StreamPosition


def create_read_all_request(options: ReadAllOptions) -> ReadReq:
    request_options = ReadReqOptions()

    request_options.resolve_links = options.resolve_links
    request_options.count = options.max_count
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())

    if options.direction == ReadDirection.FORWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Forwards
    elif options.direction == ReadDirection.BACKWARDS:
        request_options.read_direction = ReadReqOptionsReadDirection.Backwards

    request_options.all = ReadReqOptionsAllOptions()

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = ReadReqOptionsPosition()
        request_options.all.position.commit_position = (
            options.from_position.commit_position
        )
        request_options.all.position.prepare_position = (
            options.from_position.prepare_position
        )
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    if not options.filter:
        request_options.no_filter = Empty()
    else:
        request_options.filter = ReadReqOptionsFilterOptions()
        filter_expression = ReadReqOptionsFilterOptionsExpression()

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            filter_expression.regex = "^[^\\$].*"
        else:
            if options.filter.regex:
                filter_expression.regex = options.filter.regex
            if options.filter.prefix:
                filter_expression.prefix = options.filter.prefix

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, EventTypeFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, StreamNameFilter):
            request_options.filter.stream_identifier = filter_expression

        request_options.filter.max = 0
        request_options.filter.count = Empty()

    return ReadReq(options=request_options)



from __future__ import annotations

from grpclib.client import Channel

from eventstoredb.client.append_to_stream.mixin import AppendToStreamMixin
from eventstoredb.client.create_persistent_subscription_to_all.mixin import (
    CreatePersistentSubscriptionToAllMixin,
)
from eventstoredb.client.create_persistent_subscription_to_stream.mixin import (
    CreatePersistentSubscriptionToStreamMixin,
)
from eventstoredb.client.delete_persistent_subscription_to_all.mixin import (
    DeletePersistentSubscriptionToAllMixin,
)
from eventstoredb.client.delete_persistent_subscription_to_stream.mixin import (
    DeletePersistentSubscriptionToStreamMixin,
)
from eventstoredb.client.read_all.mixin import ReadAllMixin
from eventstoredb.client.read_stream.mixin import ReadStreamMixin
from eventstoredb.client.subscribe_to_all.mixin import SubscribeToAllMixin
from eventstoredb.client.subscribe_to_persistent_subscription_to_all.mixin import (
    SubscribeToPersistentSubscriptionToAllMixin,
)
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.mixin import (
    SubscribeToPersistentSubscriptionToStreamMixin,
)
from eventstoredb.client.subscribe_to_stream.mixin import SubscribeToStreamMixin
from eventstoredb.client.types import ClientOptions
from eventstoredb.client.update_persistent_subscription_to_all.mixin import (
    UpdatePersistentSubscriptionToAllMixin,
)
from eventstoredb.client.update_persistent_subscription_to_stream.mixin import (
    UpdatePersistentSubscriptionToStreamMixin,
)


class Client(
    ReadStreamMixin,
    AppendToStreamMixin,
    SubscribeToStreamMixin,
    ReadAllMixin,
    SubscribeToAllMixin,
    CreatePersistentSubscriptionToStreamMixin,
    UpdatePersistentSubscriptionToStreamMixin,
    DeletePersistentSubscriptionToStreamMixin,
    SubscribeToPersistentSubscriptionToStreamMixin,
    CreatePersistentSubscriptionToAllMixin,
    UpdatePersistentSubscriptionToAllMixin,
    DeletePersistentSubscriptionToAllMixin,
    SubscribeToPersistentSubscriptionToAllMixin,
):
    def __init__(self, options: ClientOptions | str) -> None:
        if isinstance(options, str):
            self._options = ClientOptions.from_connection_string(options)
        else:
            self._options = options
        self._channel: Channel | None = None

    @property
    def channel(self) -> Channel:
        if self._channel is None:
            self._channel = Channel(host=self._options.host, port=self._options.port)
        return self._channel



from __future__ import annotations

from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.subscribe_to_persistent_subscription_to_all.grpc import (
    create_read_request,
)
from eventstoredb.client.subscribe_to_persistent_subscription_to_all.types import (
    SubscribeToPersistentSubscriptionToAllOptions,
)
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.mixin import (
    PersistentSubscription,
)


class SubscribeToPersistentSubscriptionToAllMixin(ClientProtocol):
    def subscribe_to_persistent_subscription_to_all(
        self,
        group_name: str,
        options: SubscribeToPersistentSubscriptionToAllOptions | None = None,
    ) -> PersistentSubscription:
        if options is None:
            options = SubscribeToPersistentSubscriptionToAllOptions()

        request = create_read_request(
            group_name=group_name,
            options=options,
        )

        return PersistentSubscription(
            channel=self.channel,
            read_request=request,
        )



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

from eventstoredb.types import AllPosition, StreamRevision


class AppendExpectedRevision(Enum):
    NO_STREAM = auto()
    ANY = auto()
    STREAM_EXISTS = auto()


@dataclass
class AppendToStreamOptions:
    expected_revision: AppendExpectedRevision | StreamRevision = (
        AppendExpectedRevision.ANY
    )


@dataclass
class AppendResult:
    success: bool
    next_expected_revision: StreamRevision
    position: AllPosition | None = None



import asyncio

import pytest
from pytest_mock import MockerFixture

from eventstoredb import Client
from eventstoredb.events import JsonEvent, PersistentSubscriptionEvent
from eventstoredb.exceptions import PersistentSubscriptionNotFoundError
from eventstoredb.options import (
    CreatePersistentSubscriptionToStreamOptions,
    StreamPosition,
)

from .utils import Consumer, json_test_events


async def test_subscribe_persistent_subscription_to_stream(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_subscribe_persistent_subscription_to_stream_Test")
        ],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith("test_subscribe_persistent_subscription_to_stream_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_persistent_subscription_to_stream_Test"


async def test_subscribe_persistent_subscription_to_stream_raises_if_not_exist(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    with pytest.raises(PersistentSubscriptionNotFoundError):
        subscription = (
            eventstoredb_client.subscribe_to_persistent_subscription_to_stream(
                stream_name=stream_name,
                group_name=group_name,
            )
        )
        consumer = Consumer(subscription)
        await consumer.run_for(1)


@pytest.mark.skip(reason="test not implemented")
async def test_subscribe_persistent_subscription_to_stream_nack() -> None:
    pass


async def test_subscribe_persistent_subscription_to_stream_from_start(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
    mocker: MockerFixture,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(10),
    )
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=CreatePersistentSubscriptionToStreamOptions(
            from_revision=StreamPosition.START
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.start()
    await asyncio.sleep(1)

    assert len(consumer.events) == 10

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=json_test_events(10),
    )

    await consumer.stop(1)

    assert len(consumer.events) == 20


@pytest.mark.skip(reason="test not implemented")
async def test_subscribe_persistent_subscription_to_stream_from_end() -> None:
    pass


@pytest.mark.skip(reason="test not implemented")
async def test_subscribe_persistent_subscription_to_stream_from_revision() -> None:
    pass



from __future__ import annotations

from typing import AsyncIterator

from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.subscribe_to_stream.grpc import (
    convert_subscribe_to_stream_response,
    create_subscribe_to_stream_request,
)
from eventstoredb.client.subscribe_to_stream.types import SubscribeToStreamOptions
from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client.streams import StreamsStub

Subscription = AsyncIterator[ReadEvent]


class SubscribeToStreamMixin(ClientProtocol):
    async def subscribe_to_stream(
        self,
        stream_name: str,
        options: SubscribeToStreamOptions | None = None,
    ) -> Subscription:
        if options is None:
            options = SubscribeToStreamOptions()

        client = StreamsStub(channel=self.channel)
        request = create_subscribe_to_stream_request(
            stream_name=stream_name,
            options=options,
        )

        async for response in client.read(read_req=request):
            response_content = convert_subscribe_to_stream_response(response)
            if isinstance(response_content, ReadEvent):
                yield response_content



from __future__ import annotations

from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.subscribe_to_all.grpc import (
    convert_subscribe_to_all_response,
    create_subscribe_to_all_request,
)
from eventstoredb.client.subscribe_to_all.types import SubscribeToAllOptions
from eventstoredb.client.subscribe_to_stream.mixin import Subscription
from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client.streams import StreamsStub


class SubscribeToAllMixin(ClientProtocol):
    async def subscribe_to_all(
        self,
        options: SubscribeToAllOptions | None = None,
    ) -> Subscription:
        if options is None:
            options = SubscribeToAllOptions()

        client = StreamsStub(channel=self.channel)
        request = create_subscribe_to_all_request(options=options)

        async for response in client.read(read_req=request):
            response_content = convert_subscribe_to_all_response(response)
            if isinstance(response_content, ReadEvent):
                yield response_content



from __future__ import annotations

import betterproto

from eventstoredb.client.read_stream.grpc import convert_read_response
from eventstoredb.client.subscribe_to_stream.types import (
    SubscribeToStreamOptions,
    SubscriptionConfirmation,
)
from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.streams import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsStreamOptions,
    ReadReqOptionsSubscriptionOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespSubscriptionConfirmation,
)
from eventstoredb.types import StreamPosition, StreamRevision


def create_subscribe_to_stream_request(
    stream_name: str,
    options: SubscribeToStreamOptions,
) -> ReadReq:
    request_options = ReadReqOptions()
    request_options.resolve_links = options.resolve_links
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    request_options.subscription = ReadReqOptionsSubscriptionOptions()
    request_options.no_filter = Empty()
    request_options.stream = ReadReqOptionsStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return ReadReq(options=request_options)


def convert_subscribe_to_stream_response(
    message: ReadResp,
) -> ReadEvent | SubscriptionConfirmation:
    content_type, _ = betterproto.which_one_of(message, "content")

    if content_type == "confirmation":
        return convert_read_response_subscription_confirmation(message.confirmation)
    else:
        return convert_read_response(message)


def convert_read_response_subscription_confirmation(
    message: ReadRespSubscriptionConfirmation,
) -> SubscriptionConfirmation:
    return SubscriptionConfirmation(id=message.subscription_id)



from __future__ import annotations

from eventstoredb.client.create_persistent_subscription_to_stream.grpc import (
    create_persistent_subscription_request_settings,
)
from eventstoredb.client.update_persistent_subscription_to_stream.types import (
    UpdatePersistentSubscriptionToStreamOptions,
)
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    UpdateReq,
    UpdateReqConsumerStrategy,
    UpdateReqOptions,
    UpdateReqSettings,
    UpdateReqStreamOptions,
)
from eventstoredb.types import StreamPosition, StreamRevision


def create_update_persistent_subscription_to_stream_request(
    stream_name: str,
    group_name: str,
    options: UpdatePersistentSubscriptionToStreamOptions,
) -> UpdateReq:
    request_options = UpdateReqOptions()
    request_options.stream = UpdateReqStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())
    request_options.group_name = group_name

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    request_options.settings = create_persistent_subscription_request_settings(
        settings=options.settings,
        settings_class=UpdateReqSettings,
        consumer_strategy_class=UpdateReqConsumerStrategy,
    )

    return UpdateReq(options=request_options)



from __future__ import annotations

from dataclasses import dataclass

from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.types import AllPosition, ReadDirection, StreamPosition


@dataclass
class ReadAllOptions:
    from_position: AllPosition | StreamPosition = StreamPosition.START
    max_count: int = 2**64 - 1  # max-uint64
    direction: ReadDirection = ReadDirection.FORWARDS
    resolve_links: bool = False
    filter: ExcludeSystemEventsFilter | EventTypeFilter | StreamNameFilter | None = None



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: shared.proto
# plugin: python-betterproto
from dataclasses import dataclass

import betterproto
import betterproto.lib.google.protobuf as betterproto_lib_google_protobuf


@dataclass(eq=False, repr=False)
class Uuid(betterproto.Message):
    structured: "UuidStructured" = betterproto.message_field(1, group="value")
    string: str = betterproto.string_field(2, group="value")


@dataclass(eq=False, repr=False)
class UuidStructured(betterproto.Message):
    most_significant_bits: int = betterproto.int64_field(1)
    least_significant_bits: int = betterproto.int64_field(2)


@dataclass(eq=False, repr=False)
class Empty(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StreamIdentifier(betterproto.Message):
    stream_name: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class AllStreamPosition(betterproto.Message):
    commit_position: int = betterproto.uint64_field(1)
    prepare_position: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class WrongExpectedVersion(betterproto.Message):
    current_stream_revision: int = betterproto.uint64_field(
        1, group="current_stream_revision_option"
    )
    current_no_stream: "betterproto_lib_google_protobuf.Empty" = (
        betterproto.message_field(2, group="current_stream_revision_option")
    )
    expected_stream_position: int = betterproto.uint64_field(
        3, group="expected_stream_position_option"
    )
    expected_any: "betterproto_lib_google_protobuf.Empty" = betterproto.message_field(
        4, group="expected_stream_position_option"
    )
    expected_stream_exists: "betterproto_lib_google_protobuf.Empty" = (
        betterproto.message_field(5, group="expected_stream_position_option")
    )
    expected_no_stream: "betterproto_lib_google_protobuf.Empty" = (
        betterproto.message_field(6, group="expected_stream_position_option")
    )


@dataclass(eq=False, repr=False)
class AccessDenied(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class StreamDeleted(betterproto.Message):
    stream_identifier: "StreamIdentifier" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class Timeout(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class Unknown(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class InvalidTransaction(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class MaximumAppendSizeExceeded(betterproto.Message):
    max_append_size: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class BadRequest(betterproto.Message):
    message: str = betterproto.string_field(1)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: monitoring.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncIterator, Dict, Optional

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class StatsReq(betterproto.Message):
    use_metadata: bool = betterproto.bool_field(1)
    refresh_time_period_in_ms: int = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class StatsResp(betterproto.Message):
    stats: Dict[str, str] = betterproto.map_field(
        1, betterproto.TYPE_STRING, betterproto.TYPE_STRING
    )


class MonitoringStub(betterproto.ServiceStub):
    async def stats(
        self,
        stats_req: "StatsReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["StatsResp"]:
        async for response in self._unary_stream(
            "/event_store.client.monitoring.Monitoring/Stats",
            stats_req,
            StatsResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response


class MonitoringBase(ServiceBase):
    async def stats(self, stats_req: "StatsReq") -> AsyncIterator["StatsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_stats(
        self, stream: "grpclib.server.Stream[StatsReq, StatsResp]"
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.stats,
            stream,
            request,
        )

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.monitoring.Monitoring/Stats": grpclib.const.Handler(
                self.__rpc_stats,
                grpclib.const.Cardinality.UNARY_STREAM,
                StatsReq,
                StatsResp,
            ),
        }



from __future__ import annotations

from typing import AsyncIterator

from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.read_stream.grpc import (
    convert_read_response,
    create_read_request,
)
from eventstoredb.client.read_stream.types import ReadStreamOptions
from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client.streams import StreamsStub


class ReadStreamMixin(ClientProtocol):
    async def read_stream(
        self,
        stream_name: str,
        options: ReadStreamOptions | None = None,
    ) -> AsyncIterator[ReadEvent]:
        if options is None:
            options = ReadStreamOptions()

        client = StreamsStub(channel=self.channel)
        request = create_read_request(
            stream_name=stream_name,
            options=options,
        )

        # TODO raise exception StreamNotFoundError
        async for response in client.read(read_req=request):
            response_content = convert_read_response(response)
            if isinstance(response_content, ReadEvent):
                yield response_content



from __future__ import annotations

from eventstoredb.client.create_persistent_subscription_to_all.types import (
    CreatePersistentSubscriptionToAllOptions,
)
from eventstoredb.client.create_persistent_subscription_to_stream.grpc import (
    create_persistent_subscription_request_settings,
)
from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    CreateReq,
    CreateReqAllOptions,
    CreateReqAllOptionsFilterOptions,
    CreateReqAllOptionsFilterOptionsExpression,
    CreateReqConsumerStrategy,
    CreateReqOptions,
    CreateReqPosition,
    CreateReqSettings,
)
from eventstoredb.types import AllPosition, StreamPosition


def create_create_persistent_subscription_to_all_request(
    group_name: str,
    options: CreatePersistentSubscriptionToAllOptions,
) -> CreateReq:
    request_options = CreateReqOptions()
    request_options.group_name = group_name

    request_options.settings = create_persistent_subscription_request_settings(
        settings=options.settings,
        settings_class=CreateReqSettings,
        consumer_strategy_class=CreateReqConsumerStrategy,
    )

    request_options.all = CreateReqAllOptions()

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = CreateReqPosition()
        request_options.all.position.commit_position = (
            options.from_position.commit_position
        )
        request_options.all.position.prepare_position = (
            options.from_position.prepare_position
        )
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    if not options.filter:
        request_options.all.no_filter = Empty()
    else:
        request_options.all.filter = CreateReqAllOptionsFilterOptions()
        filter_expression = CreateReqAllOptionsFilterOptionsExpression()

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            filter_expression.regex = "^[^\\$].*"
        else:
            if options.filter.regex:
                filter_expression.regex = options.filter.regex
            if options.filter.prefix:
                filter_expression.prefix = options.filter.prefix

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            request_options.all.filter.event_type = filter_expression
        elif isinstance(options.filter, EventTypeFilter):
            request_options.all.filter.event_type = filter_expression
        elif isinstance(options.filter, StreamNameFilter):
            request_options.all.filter.stream_identifier = filter_expression

        if options.max_search_window is not None:
            request_options.all.filter.max = options.max_search_window
        else:
            request_options.all.filter.count = Empty()

        request_options.all.filter.checkpoint_interval_multiplier = (
            options.checkpoint_interval
        )

    return CreateReq(options=request_options)



from uuid import uuid4

import pytest

from eventstoredb import Client
from eventstoredb.events import JsonEvent, PersistentSubscriptionEvent
from eventstoredb.exceptions import PersistentSubscriptionNotFoundError
from eventstoredb.filters import EventTypeFilter, StreamNameFilter
from eventstoredb.options import (
    CreatePersistentSubscriptionToAllOptions,
    StreamPosition,
)

from .utils import Consumer


async def test_subscribe_persistent_subscription_to_all(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_subscribe_persistent_subscription_to_all_Test")],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith("test_subscribe_persistent_subscription_to_all_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_persistent_subscription_to_all_Test"


async def test_subscribe_persistent_subscription_to_all_raises_if_not_exist(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    with pytest.raises(PersistentSubscriptionNotFoundError):
        subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
            group_name=group_name,
        )
        consumer = Consumer(subscription)
        await consumer.run_for(1)


@pytest.mark.skip(reason="test not implemented")
async def test_subscribe_persistent_subscription_to_all_nack() -> None:
    pass


async def test_subscribe_persistent_subscription_to_all_from_start(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[
            JsonEvent(
                type="test_subscribe_persistent_subscription_to_all_from_start_Test1"
            )
        ],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.START
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[
            JsonEvent(
                type="test_subscribe_persistent_subscription_to_all_from_start_Test2"
            )
        ],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith(
            "test_subscribe_persistent_subscription_to_all_from_start_"
        )
    ]

    assert len(events) == 2

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_1
    assert (
        event.type == "test_subscribe_persistent_subscription_to_all_from_start_Test1"
    )

    assert events[1].event is not None
    event = events[1].event
    assert event.stream_name == stream_name_2
    assert (
        event.type == "test_subscribe_persistent_subscription_to_all_from_start_Test2"
    )


async def test_subscribe_persistent_subscription_to_all_from_end(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[
            JsonEvent(
                type="test_subscribe_persistent_subscription_to_all_from_end_Test1"
            )
        ],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.END
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[
            JsonEvent(
                type="test_subscribe_persistent_subscription_to_all_from_end_Test2"
            )
        ],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith(
            "test_subscribe_persistent_subscription_to_all_from_end_"
        )
    ]

    assert len(events) == 1

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_2
    assert event.type == "test_subscribe_persistent_subscription_to_all_from_end_Test2"


async def test_subscribe_persistent_subscription_to_all_from_position(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Junk")],
    )

    marker = await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Marker")],
    )
    assert marker.position is not None

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test")],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(from_position=marker.position),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.run_for(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None and not e.event.stream_name.startswith("$")
    ]

    assert len(events) == 2

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "Marker"

    assert events[1].event is not None
    event = events[1].event
    assert event.stream_name == stream_name
    assert event.type == "Test"


async def test_subscribe_persistent_subscription_to_all_filter_by_event_type_regex(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(
                type="persistent_subscription_to_all_filter_by_event_type_regex_Test1"
            ),
            JsonEvent(
                type="persistent_subscription_to_all_filter_by_event_type_regex_Test2"
            ),
        ],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            filter=EventTypeFilter(
                regex="persistent_subscription_to_all_filter_by_event_type_regex_Test1"
            )
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.run_for(1)

    assert len(consumer.events) == 1
    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name
    assert (
        event.type == "persistent_subscription_to_all_filter_by_event_type_regex_Test1"
    )


async def test_subscribe_persistent_subscription_to_all_filter_by_event_type_prefix(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(
                type="persistent_subscription_to_all_filter_by_event_type_prefix_Test1"
            ),
            JsonEvent(
                type="persistent_subscription_to_all_filter_by_event_type_prefix_Test2"
            ),
        ],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            filter=EventTypeFilter(
                prefix=["persistent_subscription_to_all_filter_by_event_type_prefix_"]
            )
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.run_for(1)

    assert len(consumer.events) == 2

    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name
    assert (
        event.type == "persistent_subscription_to_all_filter_by_event_type_prefix_Test1"
    )

    assert consumer.events[1].event is not None
    event = consumer.events[1].event
    assert event.stream_name == stream_name
    assert (
        event.type == "persistent_subscription_to_all_filter_by_event_type_prefix_Test2"
    )


async def test_subscribe_persistent_subscription_to_all_filter_by_stream_name_regex(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    stream_name_1 = (
        f"persistent_subscription_to_all_filter_by_stream_name_regex_Test1-{uuid4()}"
    )
    stream_name_2 = (
        f"persistent_subscription_to_all_filter_by_stream_name_regex_Test2-{uuid4()}"
    )

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            filter=StreamNameFilter(regex=stream_name_1)
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.run_for(1)

    assert len(consumer.events) == 1
    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"


async def test_subscribe_persistent_subscription_to_all_filter_by_stream_name_prefix(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    stream_name_1 = (
        f"persistent_subscription_to_all_filter_by_stream_name_prefix_Test1-{uuid4()}"
    )
    stream_name_2 = (
        f"persistent_subscription_to_all_filter_by_stream_name_prefix_Test2-{uuid4()}"
    )

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )

    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            filter=StreamNameFilter(
                prefix=["persistent_subscription_to_all_filter_by_stream_name_prefix_"]
            )
        ),
    )

    subscription = eventstoredb_client.subscribe_to_persistent_subscription_to_all(
        group_name=group_name,
    )

    async def on_event(event: PersistentSubscriptionEvent) -> None:
        await subscription.ack(event)

    consumer = Consumer(subscription, on_event=on_event)
    await consumer.run_for(1)

    assert len(consumer.events) == 2

    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"

    assert consumer.events[1].event is not None
    event = consumer.events[1].event
    assert event.stream_name == stream_name_2
    assert event.type == "Test"


@pytest.mark.skip(reason="test not implemented")
async def test_subscribe_persistent_subscription_to_all_exclude_System_events() -> None:
    pass



# flake8: noqa
from eventstoredb.client.append_to_stream.types import (
    AppendExpectedRevision,
    AppendToStreamOptions,
)
from eventstoredb.client.create_persistent_subscription_to_all.types import (
    CreatePersistentSubscriptionToAllOptions,
)
from eventstoredb.client.create_persistent_subscription_to_stream.types import (
    ConsumerStrategy,
    CreatePersistentSubscriptionToStreamOptions,
    PersistentSubscriptionSettings,
)
from eventstoredb.client.delete_persistent_subscription_to_all.types import (
    DeletePersistentSubscriptionToAllOptions,
)
from eventstoredb.client.delete_persistent_subscription_to_stream.types import (
    DeletePersistentSubscriptionToStreamOptions,
)
from eventstoredb.client.read_all.types import ReadAllOptions
from eventstoredb.client.read_stream.types import ReadStreamOptions
from eventstoredb.client.subscribe_to_all.types import SubscribeToAllOptions
from eventstoredb.client.subscribe_to_persistent_subscription_to_all.types import (
    SubscribeToPersistentSubscriptionToAllOptions,
)
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.types import (
    NackAction,
    SubscribeToPersistentSubscriptionToStreamOptions,
)
from eventstoredb.client.subscribe_to_stream.types import SubscribeToStreamOptions
from eventstoredb.client.types import ClientOptions
from eventstoredb.client.update_persistent_subscription_to_all.types import (
    UpdatePersistentSubscriptionToAllOptions,
)
from eventstoredb.client.update_persistent_subscription_to_stream.types import (
    UpdatePersistentSubscriptionToStreamOptions,
)
from eventstoredb.types import AllPosition, ReadDirection, StreamPosition



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto

Position = int
StreamRevision = int


class StreamPosition(Enum):
    START = auto()
    END = auto()


class ReadDirection(Enum):
    FORWARDS = auto()
    BACKWARDS = auto()


@dataclass
class AllPosition:
    commit_position: int
    prepare_position: int



# flake8: noqa
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.mixin import (
    PersistentSubscription,
)
from eventstoredb.client.subscribe_to_stream.mixin import Subscription



from uuid import uuid4

from eventstoredb import Client
from eventstoredb.events import JsonEvent
from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.options import StreamPosition, SubscribeToAllOptions

from .utils import Consumer


async def test_subscribe_to_all(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"
    it = eventstoredb_client.subscribe_to_all()

    consumer = Consumer(it)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_subscribe_to_all_Test")],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None and e.event.type.startswith("test_subscribe_to_all_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_Test"


async def test_subscribe_to_all_from_start(eventstoredb_client: Client) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    # append an event before subscription starts
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_subscribe_to_all_from_start_Test1")],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(from_position=StreamPosition.START)
    )

    consumer = Consumer(it)
    await consumer.start()

    # append an event after subscription started
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_subscribe_to_all_from_start_Test2")],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith("test_subscribe_to_all_from_start_")
    ]

    assert len(events) == 2

    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "test_subscribe_to_all_from_start_Test1"

    assert events[1].event is not None
    event = events[1].event
    assert event.stream_name == stream_name_2
    assert event.type == "test_subscribe_to_all_from_start_Test2"


async def test_subscribe_to_all_from_end(eventstoredb_client: Client) -> None:
    stream_name_1 = f"Test1-{uuid4()}"
    stream_name_2 = f"Test2-{uuid4()}"

    # append an event before subscription starts
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="test_subscribe_to_all_from_end_Test1")],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(from_position=StreamPosition.END)
    )

    consumer = Consumer(it)
    await consumer.start()

    # append an event after subscription started
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="test_subscribe_to_all_from_end_Test2")],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and e.event.type.startswith("test_subscribe_to_all_from_end_")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name_2
    assert event.type == "test_subscribe_to_all_from_end_Test2"


async def test_subscribe_to_all_from_position(eventstoredb_client: Client) -> None:
    stream_name = f"Test-{uuid4()}"

    # write some events onto stream
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Junk"), JsonEvent(type="Junk")],
    )
    # write one event to get current stream-position
    marker = await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Marker")],
    )
    assert marker.position is not None

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(from_position=marker.position)
    )

    consumer = Consumer(it)
    await consumer.run_for(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None and not e.event.type.startswith("$")
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "Test"


async def test_subscribe_to_all_exclude_system_events(
    eventstoredb_client: Client,
) -> None:
    stream_name = f"Test-{uuid4()}"

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(filter=ExcludeSystemEventsFilter())
    )

    consumer = Consumer(it)
    await consumer.start()

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[JsonEvent(type="test_subscribe_to_all_exclude_system_events_Test")],
    )

    await consumer.stop(1)

    events = [
        e
        for e in consumer.events
        if e.event is not None
        and (
            e.event.type.startswith("$")
            or e.event.type.startswith("test_subscribe_to_all_exclude_system_events_")
        )
    ]

    assert len(events) == 1
    assert events[0].event is not None
    event = events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_exclude_system_events_Test"


async def test_subscribe_to_all_filter_by_event_type_regex(
    eventstoredb_client: Client,
) -> None:
    stream_name = f"Test-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_regex_Test1"),
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_regex_Test2"),
        ],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=EventTypeFilter(
                regex="test_subscribe_to_all_filter_by_event_type_regex_Test1"
            )
        )
    )

    consumer = Consumer(it)
    await consumer.run_for(1)

    assert len(consumer.events) == 1
    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_regex_Test1"


async def test_subscribe_to_all_filter_by_event_type_prefix(
    eventstoredb_client: Client,
) -> None:
    stream_name = f"Test-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name,
        events=[
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_prefix_Test1"),
            JsonEvent(type="test_subscribe_to_all_filter_by_event_type_prefix_Test2"),
            JsonEvent(type="something_else"),
        ],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=EventTypeFilter(
                prefix=["test_subscribe_to_all_filter_by_event_type_prefix_"]
            )
        )
    )

    consumer = Consumer(it)
    await consumer.run_for(1)

    assert len(consumer.events) == 2

    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_prefix_Test1"

    assert consumer.events[1].event is not None
    event = consumer.events[1].event
    assert event.stream_name == stream_name
    assert event.type == "test_subscribe_to_all_filter_by_event_type_prefix_Test2"


async def test_subscribe_to_all_filter_by_stream_name_regex(
    eventstoredb_client: Client,
) -> None:
    stream_name_1 = f"test_subscribe_to_all_filter_by_stream_name_regex_Test1-{uuid4()}"
    stream_name_2 = f"test_subscribe_to_all_filter_by_stream_name_regex_Test2-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(filter=StreamNameFilter(regex=stream_name_1))
    )

    consumer = Consumer(it)
    await consumer.run_for(1)

    assert len(consumer.events) == 1
    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"


async def test_subscribe_to_all_filter_by_stream_name_prefix(
    eventstoredb_client: Client,
) -> None:
    stream_name_1 = (
        f"test_subscribe_to_all_filter_by_stream_name_prefix_Test1-{uuid4()}"
    )
    stream_name_2 = (
        f"test_subscribe_to_all_filter_by_stream_name_prefix_Test2-{uuid4()}"
    )
    stream_name_3 = f"Junk-{uuid4()}"

    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_1,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_2,
        events=[JsonEvent(type="Test")],
    )
    await eventstoredb_client.append_to_stream(
        stream_name=stream_name_3,
        events=[JsonEvent(type="Test")],
    )

    it = eventstoredb_client.subscribe_to_all(
        options=SubscribeToAllOptions(
            filter=StreamNameFilter(
                prefix=["test_subscribe_to_all_filter_by_stream_name_prefix_"]
            )
        )
    )

    consumer = Consumer(it)
    await consumer.run_for(1)

    assert len(consumer.events) == 2

    assert consumer.events[0].event is not None
    event = consumer.events[0].event
    assert event.stream_name == stream_name_1
    assert event.type == "Test"

    assert consumer.events[1].event is not None
    event = consumer.events[1].event
    assert event.stream_name == stream_name_2
    assert event.type == "Test"



import pytest

from eventstoredb import Client
from eventstoredb.exceptions import PersistentSubscriptionNotFoundError

from .utils import EventstoreHTTP


async def test_delete_persistent_subscription_to_all(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
    )
    subscription_before = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert subscription_before is not None
    assert subscription_before["groupName"] == group_name

    await eventstoredb_client.delete_persistent_subscription_to_all(
        group_name=group_name,
    )

    subscription_after = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert subscription_after is None


async def test_delete_persistent_subscription_to_all_raises_if_not_found(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    with pytest.raises(PersistentSubscriptionNotFoundError):
        await eventstoredb_client.delete_persistent_subscription_to_all(
            group_name=group_name,
        )



from eventstoredb.client.exceptions import ClientException


class PersistentSubscriptionError(ClientException):
    pass


class PersistentSubscriptionNotFoundError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionAlreadyExistsError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionDroppedError(PersistentSubscriptionError):
    pass


class PersistentSubscriptionMaxSubscribersReachedError(PersistentSubscriptionError):
    pass



from __future__ import annotations

from typing import AsyncIterator

from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.read_all.grpc import create_read_all_request
from eventstoredb.client.read_all.types import ReadAllOptions
from eventstoredb.client.read_stream.grpc import convert_read_response
from eventstoredb.events import ReadEvent
from eventstoredb.generated.event_store.client.streams import StreamsStub


class ReadAllMixin(ClientProtocol):
    async def read_all(
        self,
        options: ReadAllOptions | None = None,
    ) -> AsyncIterator[ReadEvent]:
        if options is None:
            options = ReadAllOptions()

        client = StreamsStub(channel=self.channel)
        request = create_read_all_request(options=options)

        # TODO raise exception StreamNotFoundError
        async for response in client.read(read_req=request):
            response_content = convert_read_response(response)
            if isinstance(response_content, ReadEvent):
                yield response_content



from __future__ import annotations

from typing import AsyncGenerator, Iterable

from eventstoredb.client.append_to_stream.grpc import (
    convert_append_response,
    create_append_header,
    create_append_request,
)
from eventstoredb.client.append_to_stream.types import (
    AppendResult,
    AppendToStreamOptions,
)
from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.events import EventData
from eventstoredb.generated.event_store.client.streams import AppendReq, StreamsStub


class AppendToStreamMixin(ClientProtocol):
    async def append_to_stream(
        self,
        stream_name: str,
        events: EventData | Iterable[EventData],
        options: AppendToStreamOptions | None = None,
    ) -> AppendResult:
        if options is None:
            options = AppendToStreamOptions()

        async def request_iterator() -> AsyncGenerator[AppendReq, None]:
            # FIXME typing of `options`
            yield create_append_header(
                stream_name=stream_name,
                options=options,  # type: ignore
            )
            if isinstance(events, EventData):
                yield create_append_request(events)
            else:
                for event in events:
                    yield create_append_request(event)

        client = StreamsStub(channel=self.channel)
        response = await client.append(request_iterator())
        return convert_append_response(stream_name, response)



from __future__ import annotations

import asyncio
from typing import AsyncIterator

from grpclib.client import Channel
from grpclib.exceptions import GRPCError

from eventstoredb.client.create_persistent_subscription_to_stream.grpc import (
    convert_grpc_error_to_exception,
)
from eventstoredb.client.protocol import ClientProtocol
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.grpc import (
    convert_read_response,
    create_ack_request,
    create_nack_request,
    create_read_request,
)
from eventstoredb.client.subscribe_to_persistent_subscription_to_stream.types import (
    NackAction,
    PersistentSubscriptionConfirmation,
    SubscribeToPersistentSubscriptionToStreamOptions,
)
from eventstoredb.events import PersistentSubscriptionEvent
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    PersistentSubscriptionsStub,
    ReadReq,
)


class SubscribeToPersistentSubscriptionToStreamMixin(ClientProtocol):
    def subscribe_to_persistent_subscription_to_stream(
        self,
        stream_name: str,
        group_name: str,
        options: SubscribeToPersistentSubscriptionToStreamOptions | None = None,
    ) -> PersistentSubscription:
        if options is None:
            options = SubscribeToPersistentSubscriptionToStreamOptions()

        request = create_read_request(
            stream_name=stream_name,
            group_name=group_name,
            options=options,
        )

        return PersistentSubscription(
            channel=self.channel,
            read_request=request,
        )


class RequestQueue(AsyncIterator[ReadReq], asyncio.Queue[ReadReq]):
    def __aiter__(self) -> AsyncIterator[ReadReq]:
        return self

    async def __anext__(self) -> ReadReq:
        return await self.get()


class PersistentSubscription(AsyncIterator[PersistentSubscriptionEvent]):
    def __init__(
        self,
        channel: Channel,
        read_request: ReadReq,
    ) -> None:
        client = PersistentSubscriptionsStub(channel=channel)

        self._request_queue = RequestQueue()
        self._request_queue.put_nowait(read_request)
        self._it = client.read(self._request_queue)

    def __aiter__(self) -> AsyncIterator[PersistentSubscriptionEvent]:
        return self

    async def __anext__(self) -> PersistentSubscriptionEvent:
        while True:
            try:
                response = await self._it.__anext__()
            except GRPCError as e:
                raise convert_grpc_error_to_exception(e)
            event = convert_read_response(response)
            if isinstance(event, PersistentSubscriptionConfirmation):
                self.id = event.id
            else:
                return event

    async def ack(
        self,
        events: PersistentSubscriptionEvent | list[PersistentSubscriptionEvent],
    ) -> None:
        request = create_ack_request(events=events)
        await self._request_queue.put(request)

    async def nack(
        self,
        action: NackAction,
        reason: str,
        events: PersistentSubscriptionEvent | list[PersistentSubscriptionEvent],
    ) -> None:
        request = create_nack_request(action=action, reason=reason, events=events)
        await self._request_queue.put(request)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: serverfeatures.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import TYPE_CHECKING, Dict, List, Optional

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class SupportedMethods(betterproto.Message):
    methods: List["SupportedMethod"] = betterproto.message_field(1)
    event_store_server_version: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class SupportedMethod(betterproto.Message):
    method_name: str = betterproto.string_field(1)
    service_name: str = betterproto.string_field(2)
    features: List[str] = betterproto.string_field(3)


class ServerFeaturesStub(betterproto.ServiceStub):
    async def get_supported_methods(
        self,
        client_empty: "__client__.Empty",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "SupportedMethods":
        return await self._unary_unary(
            "/event_store.client.server_features.ServerFeatures/GetSupportedMethods",
            client_empty,
            SupportedMethods,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class ServerFeaturesBase(ServiceBase):
    async def get_supported_methods(
        self, client_empty: "__client__.Empty"
    ) -> "SupportedMethods":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_get_supported_methods(
        self, stream: "grpclib.server.Stream[__client__.Empty, SupportedMethods]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.get_supported_methods(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.server_features.ServerFeatures/GetSupportedMethods": grpclib.const.Handler(
                self.__rpc_get_supported_methods,
                grpclib.const.Cardinality.UNARY_UNARY,
                __client__.Empty,
                SupportedMethods,
            ),
        }



import pytest

from eventstoredb import Client
from eventstoredb.exceptions import PersistentSubscriptionAlreadyExistsError
from eventstoredb.options import (
    CreatePersistentSubscriptionToStreamOptions,
    StreamPosition,
)

from .utils import EventstoreHTTP


async def test_create_persistent_subscription_to_stream(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
    )

    subscriptions = eventstoredb_httpclient.get_persistent_subscriptions(
        stream_name=stream_name,
    )
    assert len(subscriptions) == 1
    s = subscriptions[0]
    assert s["eventStreamId"] == stream_name
    assert s["groupName"] == group_name


async def test_create_persistent_subscription_to_stream_defaults(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    options = CreatePersistentSubscriptionToStreamOptions()
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=options,
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details is not None
    c = details["config"]
    s = options.settings
    assert c["startFrom"] == 0
    assert c["resolveLinktos"] == s.resolve_links
    assert c["extraStatistics"] == s.extra_statistics
    assert c["maxRetryCount"] == s.max_retry_count
    assert c["checkPointAfterMilliseconds"] == s.checkpoint_after
    assert c["minCheckPointCount"] == s.min_checkpoint_count
    assert c["maxCheckPointCount"] == s.max_checkpoint_count
    assert c["maxSubscriberCount"] == s.max_subscriber_count
    assert c["liveBufferSize"] == s.live_buffer_size
    assert c["readBatchSize"] == s.read_batch_size
    assert c["bufferSize"] == s.history_buffer_size
    assert c["namedConsumerStrategy"] == "RoundRobin"


async def test_create_persistent_subscription_to_stream_from_start(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=CreatePersistentSubscriptionToStreamOptions(
            from_revision=StreamPosition.START
        ),
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details is not None
    assert details["config"]["startFrom"] == 0


async def test_create_persistent_subscription_to_stream_from_end(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=CreatePersistentSubscriptionToStreamOptions(
            from_revision=StreamPosition.END
        ),
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details is not None
    assert details["config"]["startFrom"] == -1


async def test_create_persistent_subscription_to_stream_from_revision(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=CreatePersistentSubscriptionToStreamOptions(from_revision=123),
    )
    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details is not None
    assert details["config"]["startFrom"] == 123


async def test_create_persistent_subscription_to_stream_raises_if_exists(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
    )

    with pytest.raises(PersistentSubscriptionAlreadyExistsError):
        await eventstoredb_client.create_persistent_subscription_to_stream(
            stream_name=stream_name,
            group_name=group_name,
        )



from __future__ import annotations

from eventstoredb.client.append_to_stream.types import AppendExpectedRevision
from eventstoredb.client.exceptions import ClientException
from eventstoredb.types import StreamRevision


class AppendToStreamError(ClientException):
    pass


class StreamAlreadyExistsError(AppendToStreamError):
    def __init__(self, stream_name: str) -> None:
        self.stream_name = stream_name
        super().__init__(f"Stream '{stream_name}' already exists")


class RevisionMismatchError(AppendToStreamError):
    def __init__(
        self,
        stream_name: str,
        expected_revision: AppendExpectedRevision | StreamRevision,
        current_revision: StreamRevision | None,
    ) -> None:
        self.stream_name = stream_name
        self.expected_revision = expected_revision
        self.current_revision = current_revision
        message = (
            f"Stream '{stream_name}' has wrong revision. "
            f"(expected={expected_revision}, current={current_revision})"
        )
        super().__init__(message)



from __future__ import annotations

from dataclasses import dataclass

from yarl import URL

from eventstoredb.client.exceptions import (
    ConnectionStringMalformedError,
    ConnectionStringMissingHostError,
)


@dataclass
class ClientOptions:
    host: str
    port: int = 2113
    username: str | None = None
    password: str | None = None
    tls: bool = True
    dns_discovery: bool = False
    keep_alive_timeout: int = 10000
    keep_alive_interval: int = 10000

    @classmethod
    def from_connection_string(cls, connection_string: str) -> ClientOptions:
        try:
            url = URL(connection_string)
        except Exception:
            raise ConnectionStringMalformedError(connection_string)

        if not url.host:
            raise ConnectionStringMissingHostError(connection_string)

        options = ClientOptions(host=url.host)

        if url.scheme == "esdb":
            options.dns_discovery = False
        elif url.scheme == "esdb+discovery":
            options.dns_discovery = True

        if url.port:
            options.port = url.port

        options.username = url.user
        options.password = url.password

        return options



from __future__ import annotations

import betterproto

from eventstoredb.client.subscribe_to_all.types import Checkpoint, SubscribeToAllOptions
from eventstoredb.client.subscribe_to_stream.grpc import (
    convert_subscribe_to_stream_response,
)
from eventstoredb.client.subscribe_to_stream.types import SubscriptionConfirmation
from eventstoredb.events import ReadEvent
from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.streams import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsAllOptions,
    ReadReqOptionsFilterOptions,
    ReadReqOptionsFilterOptionsExpression,
    ReadReqOptionsPosition,
    ReadReqOptionsSubscriptionOptions,
    ReadReqOptionsUuidOption,
    ReadResp,
    ReadRespCheckpoint,
)
from eventstoredb.types import AllPosition, StreamPosition


def create_subscribe_to_all_request(options: SubscribeToAllOptions) -> ReadReq:
    request_options = ReadReqOptions()
    request_options.resolve_links = options.resolve_links
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    request_options.subscription = ReadReqOptionsSubscriptionOptions()
    request_options.all = ReadReqOptionsAllOptions()

    if isinstance(options.from_position, AllPosition):
        request_options.all.position = ReadReqOptionsPosition()
        request_options.all.position.commit_position = (
            options.from_position.commit_position
        )
        request_options.all.position.prepare_position = (
            options.from_position.prepare_position
        )
    elif options.from_position == StreamPosition.START:
        request_options.all.start = Empty()
    elif options.from_position == StreamPosition.END:
        request_options.all.end = Empty()

    if not options.filter:
        request_options.no_filter = Empty()
    else:
        request_options.filter = ReadReqOptionsFilterOptions()
        filter_expression = ReadReqOptionsFilterOptionsExpression()

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            filter_expression.regex = "^[^\\$].*"
        else:
            if options.filter.regex:
                filter_expression.regex = options.filter.regex
            if options.filter.prefix:
                filter_expression.prefix = options.filter.prefix

        if isinstance(options.filter, ExcludeSystemEventsFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, EventTypeFilter):
            request_options.filter.event_type = filter_expression
        elif isinstance(options.filter, StreamNameFilter):
            request_options.filter.stream_identifier = filter_expression

        if options.max_search_window is not None:
            request_options.filter.max = options.max_search_window
        else:
            request_options.filter.count = Empty()

        request_options.filter.checkpoint_interval_multiplier = (
            options.checkpoint_interval
        )

    return ReadReq(options=request_options)


def convert_subscribe_to_all_response(
    message: ReadResp,
) -> ReadEvent | SubscriptionConfirmation | Checkpoint:
    content_type, _ = betterproto.which_one_of(message, "content")

    if content_type == "checkpoint":
        return convert_read_response_checkpoint(message.checkpoint)
    else:
        return convert_subscribe_to_stream_response(message)


def convert_read_response_checkpoint(message: ReadRespCheckpoint) -> Checkpoint:
    return Checkpoint(
        commit_position=message.commit_position,
        prepare_position=message.prepare_position,
    )



import pytest

from eventstoredb import Client
from eventstoredb.exceptions import PersistentSubscriptionNotFoundError
from eventstoredb.options import (
    CreatePersistentSubscriptionToStreamOptions,
    StreamPosition,
    UpdatePersistentSubscriptionToStreamOptions,
)

from .utils import EventstoreHTTP


async def test_update_persistent_subscription_to_stream(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=CreatePersistentSubscriptionToStreamOptions(
            from_revision=StreamPosition.START
        ),
    )
    details_before = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details_before is not None
    assert details_before["config"]["startFrom"] == 0

    await eventstoredb_client.update_persistent_subscription_to_stream(
        stream_name=stream_name,
        group_name=group_name,
        options=UpdatePersistentSubscriptionToStreamOptions(
            from_revision=StreamPosition.END
        ),
    )

    details_after = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name=stream_name,
        group_name=group_name,
    )
    assert details_after is not None
    assert details_after["config"]["startFrom"] == -1


async def test_update_persistent_subscription_to_stream_raises_if_not_found(
    eventstoredb_client: Client,
    stream_name: str,
    group_name: str,
) -> None:
    with pytest.raises(PersistentSubscriptionNotFoundError):
        await eventstoredb_client.update_persistent_subscription_to_stream(
            stream_name=stream_name,
            group_name=group_name,
        )



from uuid import uuid4

import pytest

from eventstoredb import Client
from eventstoredb.events import JsonEvent
from eventstoredb.exceptions import PersistentSubscriptionAlreadyExistsError
from eventstoredb.filters import EventTypeFilter
from eventstoredb.options import (
    CreatePersistentSubscriptionToAllOptions,
    StreamPosition,
)

from .utils import EventstoreHTTP


async def test_create_persistent_subscription_to_all(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    assert details["eventStreamId"] == "$all"
    assert details["groupName"] == group_name


async def test_create_persistent_subscription_to_all_defaults(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    stream_name: str,
    group_name: str,
) -> None:
    options = CreatePersistentSubscriptionToAllOptions()
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=options,
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    c = details["config"]
    s = options.settings
    assert c["startFrom"] == 0
    assert c["resolveLinktos"] == s.resolve_links
    assert c["extraStatistics"] == s.extra_statistics
    assert c["maxRetryCount"] == s.max_retry_count
    assert c["checkPointAfterMilliseconds"] == s.checkpoint_after
    assert c["minCheckPointCount"] == s.min_checkpoint_count
    assert c["maxCheckPointCount"] == s.max_checkpoint_count
    assert c["maxSubscriberCount"] == s.max_subscriber_count
    assert c["liveBufferSize"] == s.live_buffer_size
    assert c["readBatchSize"] == s.read_batch_size
    assert c["bufferSize"] == s.history_buffer_size
    assert c["namedConsumerStrategy"] == "RoundRobin"


async def test_create_persistent_subscription_to_all_raises_if_exists(
    eventstoredb_client: Client,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
    )

    with pytest.raises(PersistentSubscriptionAlreadyExistsError):
        await eventstoredb_client.create_persistent_subscription_to_all(
            group_name=group_name,
        )


async def test_create_persistent_subscription_to_all_from_start(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.START
        ),
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    assert details["config"]["startPosition"] == "C:0/P:0"


async def test_create_persistent_subscription_to_all_from_end(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            from_position=StreamPosition.END
        ),
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    assert details["config"]["startPosition"] == "C:-1/P:-1"


async def test_create_persistent_subscription_to_all_from_position(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    marker = await eventstoredb_client.append_to_stream(
        stream_name=f"Test-{uuid4()}",
        events=JsonEvent(type="TestEvent"),
    )
    assert marker.position is not None
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(from_position=marker.position),
    )
    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    assert (
        details["config"]["startPosition"]
        == f"C:{marker.position.commit_position}/P:{marker.position.prepare_position}"
    )


async def test_create_persistent_subscription_to_all_filter_by_event_type_regex(
    eventstoredb_client: Client,
    eventstoredb_httpclient: EventstoreHTTP,
    group_name: str,
) -> None:
    await eventstoredb_client.create_persistent_subscription_to_all(
        group_name=group_name,
        options=CreatePersistentSubscriptionToAllOptions(
            filter=EventTypeFilter(
                regex="create_persistent_subscription_to_all_filter_by_event_type_regex"
            )
        ),
    )

    details = eventstoredb_httpclient.get_persistent_subscription_details(
        stream_name="$all",
        group_name=group_name,
    )
    assert details is not None
    # sadly we dont get the filter-settings via the http-api



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class ExcludeSystemEventsFilter:
    ...


@dataclass
class EventTypeFilter:
    regex: str | None = None
    prefix: list[str] | None = None


@dataclass
class StreamNameFilter:
    regex: str | None = None
    prefix: list[str] | None = None



from dataclasses import dataclass

from eventstoredb.client.create_persistent_subscription_to_stream.types import (
    CreatePersistentSubscriptionToStreamOptions,
)


@dataclass
class UpdatePersistentSubscriptionToStreamOptions(
    CreatePersistentSubscriptionToStreamOptions
):
    pass



from __future__ import annotations

from dataclasses import dataclass, field

from eventstoredb.client.create_persistent_subscription_to_stream.types import (
    PersistentSubscriptionSettings,
)
from eventstoredb.filters import (
    EventTypeFilter,
    ExcludeSystemEventsFilter,
    StreamNameFilter,
)
from eventstoredb.types import AllPosition, StreamPosition


@dataclass
class CreatePersistentSubscriptionToAllOptions:
    settings: PersistentSubscriptionSettings = field(
        default_factory=PersistentSubscriptionSettings
    )
    from_position: AllPosition | StreamPosition = StreamPosition.START
    filter: ExcludeSystemEventsFilter | EventTypeFilter | StreamNameFilter | None = None
    max_search_window: int | None = None
    checkpoint_interval: int = 1



from __future__ import annotations

from dataclasses import dataclass

from eventstoredb.types import StreamPosition, StreamRevision


@dataclass
class SubscribeToStreamOptions:
    from_revision: StreamRevision | StreamPosition = StreamPosition.START
    resolve_links: bool = False


@dataclass
class SubscriptionConfirmation:
    id: str



from __future__ import annotations

from eventstoredb.client.subscribe_to_persistent_subscription_to_all.types import (
    SubscribeToPersistentSubscriptionToAllOptions,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    ReadReq,
    ReadReqOptions,
    ReadReqOptionsUuidOption,
)


def create_read_request(
    group_name: str,
    options: SubscribeToPersistentSubscriptionToAllOptions,
) -> ReadReq:
    request_options = ReadReqOptions()
    request_options.all = Empty()
    request_options.group_name = group_name
    request_options.buffer_size = options.buffer_size
    request_options.uuid_option = ReadReqOptionsUuidOption(string=Empty())
    return ReadReq(options=request_options)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto


@dataclass
class SubscribeToPersistentSubscriptionToStreamOptions:
    buffer_size: int = 10


@dataclass
class PersistentSubscriptionConfirmation:
    id: str


class NackAction(Enum):
    UNKNOWN = auto()
    PARK = auto()
    RETRY = auto()
    SKIP = auto()
    STOP = auto()



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from typing import Literal
from uuid import UUID, uuid4

from eventstoredb.types import AllPosition, Position, StreamRevision


class ContentType(str, Enum):
    BINARY = "application/octet-stream"
    JSON = "application/json"


@dataclass
class EventData:
    type: str
    content_type: ContentType
    id: UUID = field(default_factory=uuid4)
    data: bytes | None = None
    metadata: bytes | None = None


@dataclass
class JsonEvent(EventData):
    content_type: Literal[ContentType.JSON] = ContentType.JSON


@dataclass
class BinaryEvent(EventData):
    content_type: Literal[ContentType.BINARY] = ContentType.BINARY


@dataclass
class RecordedEvent:
    stream_name: str
    id: UUID
    type: str
    content_type: ContentType
    revision: StreamRevision
    created: int
    position: AllPosition
    data: bytes | None
    metadata: bytes | None


@dataclass
class JsonRecordedEvent(RecordedEvent):
    ...


@dataclass
class BinaryRecordedEvent(RecordedEvent):
    ...


@dataclass
class ReadEvent:
    event: RecordedEvent | None = None
    link: RecordedEvent | None = None
    commit_position: Position | None = None


@dataclass
class PersistentSubscriptionEvent(ReadEvent):
    retry_count: int | None = None



from __future__ import annotations

from typing import Type, TypeVar

from grpclib.const import Status as GRPCStatus
from grpclib.exceptions import GRPCError

from eventstoredb.client.create_persistent_subscription_to_stream.exceptions import (
    PersistentSubscriptionAlreadyExistsError,
    PersistentSubscriptionDroppedError,
    PersistentSubscriptionError,
    PersistentSubscriptionMaxSubscribersReachedError,
    PersistentSubscriptionNotFoundError,
)
from eventstoredb.client.create_persistent_subscription_to_stream.types import (
    ConsumerStrategy,
    CreatePersistentSubscriptionToStreamOptions,
    PersistentSubscriptionSettings,
)
from eventstoredb.generated.event_store.client import Empty, StreamIdentifier
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    CreateReq,
    CreateReqConsumerStrategy,
    CreateReqOptions,
    CreateReqSettings,
    CreateReqStreamOptions,
    UpdateReqConsumerStrategy,
    UpdateReqSettings,
)
from eventstoredb.types import StreamPosition, StreamRevision

SettingsClass = TypeVar(
    "SettingsClass",
    CreateReqSettings,
    UpdateReqSettings,
)
ConsumerStrategyClass = TypeVar(
    "ConsumerStrategyClass",
    CreateReqConsumerStrategy,
    UpdateReqConsumerStrategy,
)


def create_persistent_subscription_request_settings(
    settings: PersistentSubscriptionSettings,
    settings_class: Type[SettingsClass],
    consumer_strategy_class: Type[ConsumerStrategyClass],
) -> SettingsClass:
    request_settings = settings_class()
    request_settings.resolve_links = settings.resolve_links
    request_settings.extra_statistics = settings.extra_statistics
    request_settings.max_retry_count = settings.max_retry_count
    request_settings.min_checkpoint_count = settings.min_checkpoint_count
    request_settings.max_checkpoint_count = settings.max_checkpoint_count
    request_settings.max_subscriber_count = settings.max_subscriber_count
    request_settings.live_buffer_size = settings.live_buffer_size
    request_settings.read_batch_size = settings.read_batch_size
    request_settings.history_buffer_size = settings.history_buffer_size
    request_settings.message_timeout_ms = settings.message_timeout
    request_settings.checkpoint_after_ms = settings.checkpoint_after

    if settings.consumer_strategy == ConsumerStrategy.DISPATCH_TO_SINGLE:
        request_settings.named_consumer_strategy = (
            consumer_strategy_class.DispatchToSingle  # type: ignore
        )
    elif settings.consumer_strategy == ConsumerStrategy.ROUND_ROBIN:
        request_settings.named_consumer_strategy = (
            consumer_strategy_class.RoundRobin  # type: ignore
        )
    elif settings.consumer_strategy == ConsumerStrategy.PINNED:
        request_settings.named_consumer_strategy = (
            consumer_strategy_class.Pinned  # type: ignore
        )

    return request_settings


def create_create_persistent_subscription_to_stream_request(
    stream_name: str,
    group_name: str,
    options: CreatePersistentSubscriptionToStreamOptions,
) -> CreateReq:
    request_options = CreateReqOptions()
    request_options.group_name = group_name

    request_options.settings = create_persistent_subscription_request_settings(
        settings=options.settings,
        settings_class=CreateReqSettings,
        consumer_strategy_class=CreateReqConsumerStrategy,
    )

    request_options.stream = CreateReqStreamOptions()
    request_options.stream.stream_identifier = StreamIdentifier(stream_name.encode())

    if isinstance(options.from_revision, StreamRevision):
        request_options.stream.revision = options.from_revision
    elif options.from_revision == StreamPosition.START:
        request_options.stream.start = Empty()
    elif options.from_revision == StreamPosition.END:
        request_options.stream.end = Empty()

    return CreateReq(options=request_options)


def convert_grpc_error_to_exception(
    error: GRPCError,
) -> PersistentSubscriptionError | GRPCError:
    if error.status == GRPCStatus.CANCELLED:
        return PersistentSubscriptionDroppedError(error.message)
    elif error.status == GRPCStatus.NOT_FOUND:
        return PersistentSubscriptionNotFoundError(error.message)
    elif error.status == GRPCStatus.ALREADY_EXISTS:
        return PersistentSubscriptionAlreadyExistsError(error.message)
    elif error.status == GRPCStatus.FAILED_PRECONDITION:
        return PersistentSubscriptionMaxSubscribersReachedError(error.message)
    else:
        return error



from __future__ import annotations

from eventstoredb.client.delete_persistent_subscription_to_all.types import (
    DeletePersistentSubscriptionToAllOptions,
)
from eventstoredb.generated.event_store.client import Empty
from eventstoredb.generated.event_store.client.persistent_subscriptions import (
    DeleteReq,
    DeleteReqOptions,
)


def create_delete_persistent_subscription_to_all_request(
    group_name: str,
    options: DeletePersistentSubscriptionToAllOptions,
) -> DeleteReq:
    request_options = DeleteReqOptions()

    request_options.all = Empty()
    request_options.group_name = group_name

    return DeleteReq(options=request_options)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: users.proto
# plugin: python-betterproto
from dataclasses import dataclass
from typing import TYPE_CHECKING, AsyncIterator, Dict, List, Optional

import betterproto
import grpclib
from betterproto.grpc.grpclib_server import ServiceBase

if TYPE_CHECKING:
    import grpclib.server
    from betterproto.grpc.grpclib_client import MetadataLike
    from grpclib.metadata import Deadline


@dataclass(eq=False, repr=False)
class CreateReq(betterproto.Message):
    options: "CreateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CreateReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    password: str = betterproto.string_field(2)
    full_name: str = betterproto.string_field(3)
    groups: List[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class CreateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class UpdateReq(betterproto.Message):
    options: "UpdateReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UpdateReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    password: str = betterproto.string_field(2)
    full_name: str = betterproto.string_field(3)
    groups: List[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class UpdateResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DeleteReq(betterproto.Message):
    options: "DeleteReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DeleteReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DeleteResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class EnableReq(betterproto.Message):
    options: "EnableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class EnableReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class EnableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DisableReq(betterproto.Message):
    options: "DisableReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DisableReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DisableResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class DetailsReq(betterproto.Message):
    options: "DetailsReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DetailsReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class DetailsResp(betterproto.Message):
    user_details: "DetailsRespUserDetails" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class DetailsRespUserDetails(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    full_name: str = betterproto.string_field(2)
    groups: List[str] = betterproto.string_field(3)
    last_updated: "DetailsRespUserDetailsDateTime" = betterproto.message_field(4)
    disabled: bool = betterproto.bool_field(5)


@dataclass(eq=False, repr=False)
class DetailsRespUserDetailsDateTime(betterproto.Message):
    ticks_since_epoch: int = betterproto.int64_field(1)


@dataclass(eq=False, repr=False)
class ChangePasswordReq(betterproto.Message):
    options: "ChangePasswordReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ChangePasswordReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    current_password: str = betterproto.string_field(2)
    new_password: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ChangePasswordResp(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ResetPasswordReq(betterproto.Message):
    options: "ResetPasswordReqOptions" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ResetPasswordReqOptions(betterproto.Message):
    login_name: str = betterproto.string_field(1)
    new_password: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ResetPasswordResp(betterproto.Message):
    pass


class UsersStub(betterproto.ServiceStub):
    async def create(
        self,
        create_req: "CreateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "CreateResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/Create",
            create_req,
            CreateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def update(
        self,
        update_req: "UpdateReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "UpdateResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/Update",
            update_req,
            UpdateResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def delete(
        self,
        delete_req: "DeleteReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DeleteResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/Delete",
            delete_req,
            DeleteResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def disable(
        self,
        disable_req: "DisableReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "DisableResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/Disable",
            disable_req,
            DisableResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def enable(
        self,
        enable_req: "EnableReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "EnableResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/Enable",
            enable_req,
            EnableResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def details(
        self,
        details_req: "DetailsReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> AsyncIterator["DetailsResp"]:
        async for response in self._unary_stream(
            "/event_store.client.users.Users/Details",
            details_req,
            DetailsResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        ):
            yield response

    async def change_password(
        self,
        change_password_req: "ChangePasswordReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ChangePasswordResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/ChangePassword",
            change_password_req,
            ChangePasswordResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )

    async def reset_password(
        self,
        reset_password_req: "ResetPasswordReq",
        *,
        timeout: Optional[float] = None,
        deadline: Optional["Deadline"] = None,
        metadata: Optional["MetadataLike"] = None,
    ) -> "ResetPasswordResp":
        return await self._unary_unary(
            "/event_store.client.users.Users/ResetPassword",
            reset_password_req,
            ResetPasswordResp,
            timeout=timeout,
            deadline=deadline,
            metadata=metadata,
        )


class UsersBase(ServiceBase):
    async def create(self, create_req: "CreateReq") -> "CreateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def update(self, update_req: "UpdateReq") -> "UpdateResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def delete(self, delete_req: "DeleteReq") -> "DeleteResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def disable(self, disable_req: "DisableReq") -> "DisableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def enable(self, enable_req: "EnableReq") -> "EnableResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def details(self, details_req: "DetailsReq") -> AsyncIterator["DetailsResp"]:
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def change_password(
        self, change_password_req: "ChangePasswordReq"
    ) -> "ChangePasswordResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def reset_password(
        self, reset_password_req: "ResetPasswordReq"
    ) -> "ResetPasswordResp":
        raise grpclib.GRPCError(grpclib.const.Status.UNIMPLEMENTED)

    async def __rpc_create(
        self, stream: "grpclib.server.Stream[CreateReq, CreateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.create(request)
        await stream.send_message(response)

    async def __rpc_update(
        self, stream: "grpclib.server.Stream[UpdateReq, UpdateResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.update(request)
        await stream.send_message(response)

    async def __rpc_delete(
        self, stream: "grpclib.server.Stream[DeleteReq, DeleteResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.delete(request)
        await stream.send_message(response)

    async def __rpc_disable(
        self, stream: "grpclib.server.Stream[DisableReq, DisableResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.disable(request)
        await stream.send_message(response)

    async def __rpc_enable(
        self, stream: "grpclib.server.Stream[EnableReq, EnableResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.enable(request)
        await stream.send_message(response)

    async def __rpc_details(
        self, stream: "grpclib.server.Stream[DetailsReq, DetailsResp]"
    ) -> None:
        request = await stream.recv_message()
        await self._call_rpc_handler_server_stream(
            self.details,
            stream,
            request,
        )

    async def __rpc_change_password(
        self, stream: "grpclib.server.Stream[ChangePasswordReq, ChangePasswordResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.change_password(request)
        await stream.send_message(response)

    async def __rpc_reset_password(
        self, stream: "grpclib.server.Stream[ResetPasswordReq, ResetPasswordResp]"
    ) -> None:
        request = await stream.recv_message()
        response = await self.reset_password(request)
        await stream.send_message(response)

    def __mapping__(self) -> Dict[str, grpclib.const.Handler]:
        return {
            "/event_store.client.users.Users/Create": grpclib.const.Handler(
                self.__rpc_create,
                grpclib.const.Cardinality.UNARY_UNARY,
                CreateReq,
                CreateResp,
            ),
            "/event_store.client.users.Users/Update": grpclib.const.Handler(
                self.__rpc_update,
                grpclib.const.Cardinality.UNARY_UNARY,
                UpdateReq,
                UpdateResp,
            ),
            "/event_store.client.users.Users/Delete": grpclib.const.Handler(
                self.__rpc_delete,
                grpclib.const.Cardinality.UNARY_UNARY,
                DeleteReq,
                DeleteResp,
            ),
            "/event_store.client.users.Users/Disable": grpclib.const.Handler(
                self.__rpc_disable,
                grpclib.const.Cardinality.UNARY_UNARY,
                DisableReq,
                DisableResp,
            ),
            "/event_store.client.users.Users/Enable": grpclib.const.Handler(
                self.__rpc_enable,
                grpclib.const.Cardinality.UNARY_UNARY,
                EnableReq,
                EnableResp,
            ),
            "/event_store.client.users.Users/Details": grpclib.const.Handler(
                self.__rpc_details,
                grpclib.const.Cardinality.UNARY_STREAM,
                DetailsReq,
                DetailsResp,
            ),
            "/event_store.client.users.Users/ChangePassword": grpclib.const.Handler(
                self.__rpc_change_password,
                grpclib.const.Cardinality.UNARY_UNARY,
                ChangePasswordReq,
                ChangePasswordResp,
            ),
            "/event_store.client.users.Users/ResetPassword": grpclib.const.Handler(
                self.__rpc_reset_password,
                grpclib.const.Cardinality.UNARY_UNARY,
                ResetPasswordReq,
                ResetPasswordResp,
            ),
        }
